# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)

c1 = global_namespace.variable("c1")
print(str(c1), type(c1))
# > 'c1 [variable]', <class 'pygccxml.declarations.variable.variable_t'>

print(str(c1.decl_type), type(c1.decl_type))
# > 'int const', <class 'pygccxml.declarations.cpptypes.const_t'>

base = declarations.remove_const(c1.decl_type)
print(str(base), type(base))
# > 'int', <class 'pygccxml.declarations.cpptypes.int_t'>

c2 = global_namespace.variable("c2")
print(str(c2.decl_type), type(c2.decl_type))
# > 'int const', <class 'pygccxml.declarations.cpptypes.const_t'>
# Even if the declaration was defined as 'const int', pygccxml will always
# output the const qualifier (and some other qualifiers) on the right hand
# side (by convention).

cv1 = global_namespace.variable("cv1")
print(str(cv1.decl_type), type(cv1.decl_type))
# > 'int const volatile', <class 'pygccxml.declarations.cpptypes.volatile_t'>

# Remove one level:
base = declarations.remove_volatile(cv1.decl_type)
print(str(base), type(base))
# > 'int const', <class 'pygccxml.declarations.cpptypes.const_t'>

# Remove the second level:
base = declarations.remove_const(base)
print(str(base), type(base))
# > 'int', <class 'pygccxml.declarations.cpptypes.int_t'>

# We can also directly do this in one step:
base = declarations.remove_cv(cv1.decl_type)
print(str(base), type(base))
# > 'int', <class 'pygccxml.declarations.cpptypes.int_t'>

# As for consts, the const and volatile are on the right hand side
# (by convention), and always in the same order
cv2 = global_namespace.variable("cv2")
print(str(cv2.decl_type), type(cv2.decl_type))
# > 'int const volatile', <class 'pygccxml.declarations.cpptypes.volatile_t'>

# And a last example with a pointer_t:
cptr1 = global_namespace.variable("cptr1")
print(str(cptr1.decl_type), type(cptr1.decl_type))
# > 'int const * const', <class 'pygccxml.declarations.cpptypes.const_t'>)

base = declarations.remove_const(cptr1.decl_type)
print(str(base), type(base))
# > 'int const *', <class 'pygccxml.declarations.cpptypes.pointer_t'>

base = declarations.remove_pointer(base)
print(str(base), type(base))
# > 'int const', <class 'pygccxml.declarations.cpptypes.const_t'>)

base = declarations.remove_const(base)
print(str(base), type(base))
# > 'int', <class 'pygccxml.declarations.cpptypes.int_t'>)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ["has_public_binary_operator_traits.hpp"]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_yes_equal(global_ns):
    yes_ns = global_ns.namespace('yesequal')
    for typedef in yes_ns.typedefs():
        assert declarations.has_public_equal(typedef) is True


def test_no_equal(global_ns):
    no_ns = global_ns.namespace('noequal')
    for typedef in no_ns.typedefs():
        assert declarations.has_public_equal(typedef) is False


def test_yes_less(global_ns):
    yes_ns = global_ns.namespace('yesless')
    for typedef in yes_ns.typedefs():
        assert declarations.has_public_less(typedef)


def test_no_less(global_ns):
    no_ns = global_ns.namespace('noless')
    for typedef in no_ns.typedefs():
        assert declarations.has_public_less(typedef) is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "declarations_calldef.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


TEMPLATE = """
    //test generated declaration string using gcc(xml) compiler
    #include "declarations_calldef.hpp"
    void test_generated_decl_string( %s );
    """


def test_member_function(global_ns):
    config = autoconfig.cxx_parsers_cfg.config.clone()
    member_inline_call = \
        global_ns.member_function('member_inline_call')
    decls = parser.parse_string(
        TEMPLATE %
        member_inline_call.decl_string,
        config)
    assert decls is not None


def test_free_function(global_ns):
    config = autoconfig.cxx_parsers_cfg.config.clone()
    return_default_args = \
        global_ns.free_function('return_default_args')
    decls = parser.parse_string(
        TEMPLATE %
        return_default_args.decl_string,
        config)
    assert decls is not None


def test_all_mem_and_free_funs(global_ns):
    config = autoconfig.cxx_parsers_cfg.config.clone()
    ns = global_ns.namespace('::declarations::calldef')
    for f in ns.member_functions():
        decls = parser.parse_string(
            TEMPLATE % f.decl_string, config)
        assert decls is not None
    for f in ns.free_functions():
        decls = parser.parse_string(
            TEMPLATE % f.decl_string, config)
        assert decls is not None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "declarations_calldef.hpp",
]


@pytest.fixture
def decls():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    return decls


def test_calldef_matcher(decls):
    criteria = declarations.calldef_matcher_t(
        name='return_default_args',
        return_type='int',
        arg_types=[None, declarations.bool_t()])
    rda = declarations.matcher.get_single(criteria, decls)
    assert rda is not None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = ['find_noncopyable_vars.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_find_noncopyable_vars(global_ns):
    """
    Test the find_noncopyable_vars function

    """

    # The ptr1 variable in the holder struct can be copied,
    # but not the ptr2 variable
    holder = global_ns.class_("holder")
    nc_vars = declarations.find_noncopyable_vars(holder)
    assert len(nc_vars) == 1
    assert nc_vars[0].name == "ptr2"
    assert declarations.is_pointer(nc_vars[0].decl_type) is True
    assert declarations.is_const(nc_vars[0].decl_type) is True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find the location of the xml generator (castxml or gccxml)
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

# Parse the c++ file
decls = parser.parse([filename], xml_generator_config)

# Get access to the global namespace
global_namespace = declarations.get_global_namespace(decls)

# Get access to the 'ns' namespace
ns = global_namespace.namespace("ns")



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


class xml_generators(object):
    """
    Helper class for the xml generator versions
    """
    __gccxml_06 = "GCC-XML 0.6"
    __gccxml_07 = "GCC-XML 0.7"
    __gccxml_09 = "GCC-XML 0.9"
    __gccxml_09_buggy = "GCC-XML 0.9 BUGGY"
    __castxml = "CastXML (gccxml like)"
    __castxml1 = "CastXML (format version 1)"
    __separator = "@"

    def __init__(self, logger, gccxml_cvs_revision=None, castxml_format=None):
        """
        Create a new xml_generators object.

        Args:
            logger (logging.Logger) : a logger for debugging output
            gccxml_cvs_revision (str|None): the xml output version
            castxml_format (str|None): the xml output version
        """

        if castxml_format is not None and gccxml_cvs_revision is not None:
            raise RuntimeError("Setting both gccxml_cvs_revision and"
                               "castxml_format is not allowed!")

        self._is_castxml1 = False
        self._is_castxml = False
        self._is_gccxml = False

        if castxml_format is not None:
            self._xml_generator_version = self.__castxml
            self._xml_output_version = castxml_format
            self._is_castxml = True
            self._is_castxml1 = True
        elif gccxml_cvs_revision is not None:
            self._xml_generator_version, self._xml_output_version = \
                self.__extract_versions(logger, gccxml_cvs_revision)
            self._is_gccxml = "GCC-XML" in self._xml_generator_version
            self._is_castxml = "CastXML" in self._xml_generator_version
        else:
            raise RuntimeError("Either castxml_format or gccxml_cvs_revision"
                               "need to be defined!")

    def __extract_versions(self, logger, xml_output_version):
        if xml_output_version is None or xml_output_version == "0.6":
            # It is not clear what gccxml 0.6 had as version (0.6 or None),
            # but both cases get caught here.
            logger.debug("GCCXML version - 0.6")
            xml_generator = self.__gccxml_06
            xml_output_version = 0.6
        else:
            xml_output_version = float(xml_output_version)
            if xml_output_version <= 1.114:
                logger.debug("GCCXML version - 0.7")
                xml_generator = self.__gccxml_07
            elif 1.115 <= xml_output_version <= 1.126:
                logger.debug(
                    "GCCXML version - 0.9 BUGGY ( %s )", xml_output_version)
                xml_generator = self.__gccxml_09_buggy
            elif 1.127 <= xml_output_version <= 1.135:
                logger.debug(
                    "GCCXML version - 0.9 ( %s )", xml_output_version)
                xml_generator = self.__gccxml_09
            else:
                # CastXML starts with revision 1.136, but still writes the
                # GCCXML tag and the 0.9 version number in the XML files
                # for backward compatibility.
                logger.debug(
                    "CASTXML version - None ( %s )", xml_output_version)
                xml_generator = self.__castxml
        return xml_generator, str(xml_output_version)

    def get_string_repr(self):
        """
        Get a string identifier for the current type of xml generator

        Returns:
            str: identifier
        """
        return \
            self._xml_generator_version + \
            self.__separator + \
            str(self._xml_output_version)

    @property
    def is_gccxml(self):
        """
        Is the current xml generator gccxml?

        Returns:
            bool: is gccxml being used?
        """
        return self._is_gccxml

    @property
    def is_castxml(self):
        """
        Is the current xml generator castxml?

        Returns:
            bool: is castxml being used?
        """
        return self._is_castxml

    @property
    def is_castxml1(self):
        """
        Is the current xml generator castxml (with output format version 1)?

        Returns:
            bool: is castxml (with output format version 1) being used?
        """
        return self._is_castxml1

    @property
    def is_gccxml_06(self):
        """
        Is the current xml generator gccxml (version 0.6)?

        Returns:
            bool: is gccxml 0.6 being used?
        """
        return self._xml_generator_version == self.__gccxml_06

    @property
    def is_gccxml_07(self):
        """
        Is the current xml generator gccxml (version 0.7)?

        Returns:
            bool: is gccxml 0.7 being used?
        """
        return self._xml_generator_version == self.__gccxml_07

    @property
    def is_gccxml_09(self):
        """
        Is the current xml generator gccxml (version 0.9)?

        Returns:
            bool: is gccxml 0.9 being used?
        """
        return self._xml_generator_version == self.__gccxml_09

    @property
    def is_gccxml_09_buggy(self):
        """
        Is the current xml generator gccxml (version 0.9 - buggy)?

        Returns:
            bool: is gccxml 0.9 (buggy) being used?
        """
        return self._xml_generator_version == self.__gccxml_09_buggy

    @property
    def xml_output_version(self):
        """
        The current xml output version for the parsed file.

        Returns:
            str: the xml output version
        """
        return self._xml_output_version



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import declarations


class linker_t(
        declarations.decl_visitor_t,
        declarations.type_visitor_t,
        object):

    def __init__(
            self, decls, types, access, membership,
            files, xml_generator_from_xml_file=None):
        declarations.decl_visitor_t.__init__(self)
        declarations.type_visitor_t.__init__(self)
        object.__init__(self)

        self.__decls = decls
        self.__types = types
        self.__access = access
        self.__membership = membership
        self.__files = files
        self.__inst = None
        self.__xml_generator_from_xml_file = xml_generator_from_xml_file

    @property
    def instance(self):
        return self.__inst

    @instance.setter
    def instance(self, inst):
        """
        Called by __parse_xml_file in source_reader.

        """

        self.__inst = inst

        # use inst, to reduce attribute access time
        if isinstance(inst, declarations.declaration_t) and \
                inst.location is not None and \
                inst.location.file_name != '':
            inst.location.file_name = self.__files[inst.location.file_name]

    def __link_type(self, type_id):
        if type_id is None:
            # in some situations type_id is None, return_type of constructor or
            # destructor
            return None
        elif type_id in self.__types:
            return self.__types[type_id]
        elif type_id in self.__decls:
            base = declarations.declarated_t(declaration=self.__decls[type_id])
            self.__types[type_id] = base
            return base
        elif type_id == '...':
            return declarations.ellipsis_t()

        return declarations.unknown_t()

    def __link_compound_type(self):
        self.__inst.base = self.__link_type(self.__inst.base)

    def __link_members(self):
        if id(self.__inst) not in self.__membership:
            return
        for member in self.__membership[id(self.__inst)]:
            if member not in self.__access:
                continue
            access = self.__access[member]
            if member not in self.__decls:
                continue
            decl = self.__decls[member]
            if isinstance(self.__inst, declarations.class_t):
                self.__inst.adopt_declaration(decl, access)
            else:
                self.__inst.adopt_declaration(decl)

    def __link_calldef(self):
        self.__inst.return_type = self.__link_type(self.__inst.return_type)
        if isinstance(self.__inst, declarations.type_t):
            linked_args = [
                self.__link_type(arg) for arg in self.__inst.arguments_types]
            self.__inst.arguments_types = linked_args
        else:
            for arg in self.__inst.arguments:
                arg.decl_type = self.__link_type(arg.decl_type)
            for i, exception in enumerate(self.__inst.exceptions):
                try:
                    self.__inst.exceptions[i] = self.__decls[exception]
                except KeyError:
                    self.__inst.exceptions[i] = self.__link_type(exception)

    def visit_member_function(self):
        self.__link_calldef()

    def visit_constructor(self):
        self.__link_calldef()

    def visit_destructor(self):
        self.__link_calldef()

    def visit_member_operator(self):
        self.__link_calldef()

    def visit_comment(self):
        pass

    def visit_casting_operator(self):
        self.__link_calldef()
        # FIXME: is the patch still needed as the demangled name support has
        # been dropped?
        # will be fixed by patcher. It is needed because of demangled name
        # taken into account
        # self.__inst._name = 'operator ' + self.__inst.return_type.decl_string

    def visit_free_function(self):
        self.__link_calldef()

    def visit_free_operator(self):
        self.__link_calldef()

    def visit_class_declaration(self):
        pass

    def visit_class(self):
        self.__link_members()
        # GCC-XML sometimes generates constructors with names that does not
        # match class name. I think this is because those constructors are
        # compiler generated. I need to find out more about this and to talk
        # with Brad

        new_name = self.__inst._name
        if declarations.templates.is_instantiation(new_name):
            new_name = declarations.templates.name(new_name)

        for decl in self.__inst.declarations:
            if not isinstance(decl, declarations.constructor_t):
                continue
            if '.' in decl._name or '$' in decl._name:
                decl._name = new_name

        bases = self.__inst.bases.split()
        self.__inst.bases = []
        for base in bases:
            # it could be "_5" or "protected:_5"
            data = base.split(':')
            base_decl = self.__decls[data[-1]]
            access = declarations.ACCESS_TYPES.PUBLIC
            if len(data) == 2:
                access = data[0]
            self.__inst.bases.append(
                declarations.hierarchy_info_t(base_decl, access))
            base_decl.derived.append(
                declarations.hierarchy_info_t(self.__inst, access))

    def visit_ellipsis(self):
        pass

    def visit_enumeration(self):
        pass

    def visit_namespace(self):
        self.__link_members()

    def visit_typedef(self):
        self.__inst.decl_type = self.__link_type(self.__inst.decl_type)

    def visit_variable(self):
        self.__inst.decl_type = self.__link_type(self.__inst.decl_type)

    def visit_void(self):
        pass

    def visit_char(self):
        pass

    def visit_signed_char(self):
        pass

    def visit_unsigned_char(self):
        pass

    def visit_wchar(self):
        pass

    def visit_short_int(self):
        pass

    def visit_short_unsigned_int(self):
        pass

    def visit_bool(self):
        pass

    def visit_int(self):
        pass

    def visit_unsigned_int(self):
        pass

    def visit_long_int(self):
        pass

    def visit_long_unsigned_int(self):
        pass

    def visit_long_long_int(self):
        pass

    def visit_long_long_unsigned_int(self):
        pass

    def visit_int128(self):
        pass

    def visit_uint128(self):
        pass

    def visit_float(self):
        pass

    def visit_double(self):
        pass

    def visit_long_double(self):
        pass

    def visit_complex_long_double(self):
        pass

    def visit_complex_double(self):
        pass

    def visit_complex_float(self):
        pass

    def visit_jbyte(self):
        pass

    def visit_jshort(self):
        pass

    def visit_jint(self):
        pass

    def visit_jlong(self):
        pass

    def visit_jfloat(self):
        pass

    def visit_jdouble(self):
        pass

    def visit_jchar(self):
        pass

    def visit_jboolean(self):
        pass

    def visit_volatile(self):
        if isinstance(self.__inst.base, declarations.const_t):
            const_type_inst = self.__inst.base
            const_type_inst.base = self.__link_type(const_type_inst.base)
        else:
            self.__link_compound_type()

    def visit_const(self):
        self.__link_compound_type()

    def visit_pointer(self):
        if isinstance(self.__inst.base, declarations.member_variable_type_t):
            original_inst = self.__inst
            self.__inst = self.__inst.base
            self.visit_member_variable_type()
            self.__inst = original_inst
        else:
            self.__link_compound_type()

    def visit_reference(self):
        self.__link_compound_type()

    def visit_elaborated(self):
        self.__link_compound_type()

    def visit_array(self):
        self.__link_compound_type()

    def visit_free_function_type(self):
        self.__link_calldef()

    def visit_member_function_type(self):
        self.__link_calldef()
        if isinstance(self.__inst, declarations.type_t):
            self.__inst.class_inst = self.__link_type(self.__inst.class_inst)

    def visit_member_variable_type(self):
        self.__inst.variable_type = self.__link_type(self.__inst.variable_type)
        self.__link_compound_type()

    def visit_declarated(self):
        if isinstance(self.__inst.declaration, str):
            self.__inst.declaration = self.__decls[self.__inst.declaration]

    def visit_restrict(self):
        self.__link_compound_type()



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import inspect

from pygccxml import declarations


def test_types_hashes():
    """
    Test if all the type_t instances implement a hash method.

    The hash is part of the public API, as there are multiple tools
    that rely on it to compare type_t instances.

    The best way to test this is to instanciate dummy type_t objects
    for each class that subclasses type_t, and check that the hash of
    these objects is not None.

    """

    members = inspect.getmembers(declarations, inspect.isclass)
    for member in members:
        member_type = member[1]
        is_type_t_subclass = issubclass(member_type, declarations.type_t)
        is_not_type_t = member_type != declarations.type_t
        if is_type_t_subclass and is_not_type_t:
            type_mockup = _create_type_t_mockup(member_type)
            assert hash(type_mockup) is not None


def test_declarations_hashes():
    """
    Test if all the declaration_t instances implement a hash method.

    The hash is part of the public API, as there are multiple tools
    that rely on it to compare declaration_t instances.

    The best way to test this is to instanciate dummy declaration_t objects
    for each class that subclasses declaration_t, and check that the hash
    of these objects is not None.

    """
    members = inspect.getmembers(declarations, inspect.isclass)
    for member in members:
        member_type = member[1]
        if issubclass(member_type, declarations.declaration_t):
            assert hash(member_type()) is not None


def test_type_qualifiers_t_hash():
    """
    Test if the type_qualifiers_t instance implements a hash method.

    The hash is part of the public API, as there are multiple tools
    that rely on it to compare type_qualifiers_t instances.

    """
    assert hash(declarations.type_qualifiers_t()) is not None


def _create_type_t_mockup(member_type):
    nbr_parameters = len(inspect.signature(member_type).parameters)
    if nbr_parameters == 0:
        m = member_type()
    else:
        if member_type == declarations.array_t:
            m = member_type(_base_mockup(), size=1)
        else:
            m = member_type(_base_mockup())
    m.cache.decl_string = ""
    return m


class _base_mockup(declarations.type_t):

    def __init__(self):
        declarations.type_t.__init__(self)
        self.cache.decl_string = ""
        self._decl_string = ""
        self.variable_type = declarations.type_t()

    def build_decl_string(self, with_defaults=False):
        return self._decl_string



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import sys

# This wrapper layer allows to execute the examples for the unittests
# in a separate environment, where the sys.path has been hacked to point
# to the pygccxml source tree (we do not want to pickup any pygccxml installed
# on the system). Unfortunately passing the PYTHONPATH as env variable to
# subprocess just appends to the sys.path ... here we want to insert at
# position 1 so that the right pygccxml is found


example_file = sys.argv[1]
example_folder = os.path.dirname(example_file)
root = os.path.normpath(
    os.path.dirname(sys.modules[__name__].__file__) + "/../")

# Add the root folder to the path, so that the examples are not run against
# whatever pygccxml version is installed
sys.path.insert(1, root)

# Hack the __file__ location, so that the example thinks it is in the
# docs/examples folder
sys.modules[__name__].__file__ = example_file

# Run the example
with open(example_file) as f:
    code = compile(f.read(), example_file, "exec")
    exec(code, None, None)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import warnings
warnings.simplefilter("error", Warning)

# Find the location of the xml generator (castxml or gccxml)
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# Write a string containing some c++ code
code = """
    class MyClass {
        int a;
    };
"""

# Parse the code
decls = parser.parse_string(code, xml_generator_config)

# Get access to the global namespace
global_ns = declarations.get_global_namespace(decls)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import re
from .. import utils
from .. import declarations
from ..declarations import type_traits


class default_argument_patcher_t(object):

    def __init__(self, enums, cxx_std):
        object.__init__(self)
        self.__enums = enums
        self.__cxx_std = cxx_std

    def __call__(self, decl):
        for arg in decl.arguments:
            if not arg.default_value:
                continue
            fixer = self.__find_fixer(decl, arg)
            if fixer:
                arg.default_value = fixer(decl, arg)

    def __find_fixer(self, func, arg):
        if self.__is_unqualified_enum(func, arg):
            return self.__fix_unqualified_enum
        elif self.__is_double_call(func, arg):
            return self.__fix_double_call
        elif self.__is_invalid_integral(func, arg):
            return self.__fix_invalid_integral
        elif self.__is_constructor_call(func, arg):
            return self.__fix_constructor_call

        return None

    @staticmethod
    def __join_names(prefix, suffix):
        if prefix == '::':
            return '::' + suffix

        return prefix + '::' + suffix

    @staticmethod
    def __is_unqualified_enum(func, arg):
        type_ = declarations.remove_reference(
            declarations.remove_cv(arg.decl_type))
        if not declarations.is_enum(type_):
            return False
        enum_type = declarations.enum_declaration(type_)
        # GCCXML does not qualify an enum value in the default argument
        # but CastXML does. Split the default value and use only the
        # enum value for fixing it.
        return enum_type.has_value_name(
            arg.default_value.split('::')[-1])

    def __fix_unqualified_enum(self, func, arg):
        type_ = declarations.remove_reference(
            declarations.remove_cv(arg.decl_type))
        enum_type = declarations.enum_declaration(type_)
        if self.__cxx_std.is_cxx11_or_greater:
            qualifier_decl_string = enum_type.decl_string
        else:
            qualifier_decl_string = enum_type.parent.decl_string
        return self.__join_names(
            qualifier_decl_string,
            arg.default_value.split('::')[-1])

    @staticmethod
    def __is_invalid_integral(func, arg):
        type_ = declarations.remove_reference(
            declarations.remove_cv(arg.decl_type))
        if not declarations.is_integral(type_):
            return False
        try:
            int(arg.default_value)
            return False
        except ValueError:
            # The arg.default_value string could not be converted to int
            return True

    def __fix_invalid_integral(self, func, arg):
        try:
            int(arg.default_value)
            return arg.default_value
        except ValueError:
            # The arg.default_value string could not be converted to int
            pass

        try:
            int(arg.default_value, 16)
            if utils.get_architecture() == 64:
                # on 64 bit architecture, gccxml reports 0fffff, which is
                # valid number the problem is that in this case it is so
                # buggy so pygccxml can not fix it users will have to fix the
                # default value manually
                return arg.default_value
            default_value = arg.default_value.lower()
            found_hex = [ch for ch in default_value if ch in 'abcdef']
            if found_hex and not default_value.startswith('0x'):
                int('0x' + default_value, 16)
                return '0x' + default_value
        except ValueError:
            # The arg.default_value string could not be converted to int
            pass

        # may be we deal with enum
        # CastXML qualifies the enum value with enum type, so split the
        # argument and use only the enum value
        enum_value = arg.default_value.split('::')[-1]
        parent = func.parent
        while parent:
            found = self.__find_enum(parent, enum_value)
            if found:
                if declarations.is_fundamental(arg.decl_type) and ' ' in \
                        arg.decl_type.decl_string:
                    template = '(%s)(%s)'
                else:
                    template = '%s(%s)'
                if self.__cxx_std.is_cxx11_or_greater:
                    qualifier_decl_string = found.decl_string
                else:
                    qualifier_decl_string = found.parent.decl_string
                return template % (arg.decl_type.decl_string,
                                   self.__join_names(qualifier_decl_string,
                                                     enum_value))
            else:
                parent = parent.parent

        # check if we have an unqualified integral constant
        # only do patching in cases where we have a bare variable name
        c_var = re.compile("[a-z_][a-z0-9_]*", re.IGNORECASE)
        m = c_var.match(arg.default_value)
        if m:
            parent = func.parent
            while parent:
                try:
                    found = parent.variable(
                        arg.default_value, recursive=False)
                except declarations.declaration_not_found_t:
                    # ignore exceptions if a match is not found
                    found = None
                if found and declarations.is_fundamental(arg.decl_type):
                    return "%s" % self.__join_names(
                        found.parent.decl_string, arg.default_value)
                parent = parent.parent

        return arg.default_value

    def __find_enum(self, scope, default_value):
        # this algorithm could be improved: it could take into account
        # 1. unnamed namespace
        # 2. location within files

        for enum in self.__enums:
            if enum.parent is scope and enum.has_value_name(default_value):
                return enum
        return None

    @staticmethod
    def __is_double_call(func, arg):
        call_invocation = declarations.call_invocation
        dv = arg.default_value
        found1 = call_invocation.find_args(dv)
        if found1 == call_invocation.NOT_FOUND:
            return False
        found2 = call_invocation.find_args(dv, found1[1] + 1)
        if found2 == call_invocation.NOT_FOUND:
            return False
        args1 = call_invocation.args(dv[found1[0]: found1[1] + 1])
        args2 = call_invocation.args(dv[found2[0]: found2[1] + 1])
        return len(args1) == len(args2)

    @staticmethod
    def __fix_double_call(func, arg):
        call_invocation = declarations.call_invocation
        dv = arg.default_value
        found1 = call_invocation.find_args(dv)
        found2 = call_invocation.find_args(dv, found1[1] + 1)
        # args1 = call_invocation.args(dv[ found1[0] : found1[1] + 1 ])
        args2 = call_invocation.args(dv[found2[0]: found2[1] + 1])
        return call_invocation.join(dv[:found1[0]], args2)

    @staticmethod
    def __is_constructor_call(func, arg):
        call_invocation = declarations.call_invocation
        dv = arg.default_value
        if not call_invocation.is_call_invocation(dv):
            return False
        name = call_invocation.name(dv)
        base_type = type_traits.base_type(arg.decl_type)
        if not isinstance(base_type, declarations.declarated_t):
            return False
        decl = base_type.declaration
        return (
            decl.name == name or
            (isinstance(decl, declarations.class_t) and
             default_argument_patcher_t.__is_decl_in_aliases(decl, name)))

    @staticmethod
    def __is_decl_in_aliases(declaration, name):
        return name in [typedef.name for typedef in declaration.aliases]

    def __fix_constructor_call(self, func, arg):
        call_invocation = declarations.call_invocation
        dv = arg.default_value
        if not call_invocation.is_call_invocation(dv):
            return False
        base_type = type_traits.base_type(arg.decl_type)
        decl = base_type.declaration
        name, args = call_invocation.split(dv)
        if decl.name != name:
            # we have some alias to the class
            relevant_typedefs = [
                typedef for typedef in decl.aliases if typedef.name == name]
            if len(relevant_typedefs) == 1:
                f_q_name = self.__join_names(
                    declarations.full_name(
                        relevant_typedefs[0].parent),
                    name)
            else:  # in this case we can not say which typedef user uses:
                f_q_name = self.__join_names(
                    declarations.full_name(
                        decl.parent),
                    decl.name)
        else:
            f_q_name = self.__join_names(
                declarations.full_name(
                    decl.parent),
                name)

        return call_invocation.join(f_q_name, args)


class casting_operator_patcher_t(object):

    def __init__(self):
        object.__init__(self)

    def __call__(self, decl):
        decl.name = 'operator ' + decl.return_type.decl_string


_casting_oper_patcher_ = casting_operator_patcher_t()


def fix_calldef_decls(decls, enums, cxx_std):
    """
    some times gccxml report typedefs defined in no namespace
    it happens for example in next situation
    template< typename X>
    void ddd(){ typedef typename X::Y YY;}
    if I will fail on this bug next time, the right way to fix it may be
    different
    """
    default_arg_patcher = default_argument_patcher_t(enums, cxx_std)
    # decls should be flat list of all declarations, you want to apply patch on
    for decl in decls:
        default_arg_patcher(decl)
        if isinstance(decl, declarations.casting_operator_t):
            _casting_oper_patcher_(decl)


def update_unnamed_class(decls):
    """
    Adds name to class_t declarations.

    If CastXML is being used, the type definitions with an unnamed
    class/struct are split across two nodes in the XML tree. For example,

        typedef struct {} cls;

    produces

        <Struct id="_7" name="" context="_1" .../>
        <Typedef id="_8" name="cls" type="_7" context="_1" .../>

    For each typedef, we look at which class it refers to, and update the name
    accordingly. This helps the matcher classes finding these declarations.
    This was the behaviour with gccxml too, so this is important for
    backward compatibility.

    If the castxml epic version 1 is used, there is even an elaborated type
    declaration between the typedef and the struct/class, that also needs to be
    taken care of.

    Args:
        decls (list[declaration_t]): a list of declarations to be patched.
    Returns:
        None
    """

    for decl in decls:
        if isinstance(decl, declarations.typedef_t):
            referent = decl.decl_type
            if isinstance(referent, declarations.elaborated_t):
                referent = referent.base
            if not isinstance(referent, declarations.declarated_t):
                continue
            referent = referent.declaration
            if referent.name or not isinstance(referent, declarations.class_t):
                continue
            referent.name = decl.name



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import pytest

from . import autoconfig

from pygccxml import parser


def test_raise():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    content = "abra cadabra " + os.linesep
    with pytest.raises(RuntimeError) as _:
        parser.parse_string(content, config)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines declarations visitor class interface
"""


class decl_visitor_t(object):

    """
    declarations visitor interface

    All functions within this class should be redefined in derived classes.
    """

    def __init__(self):
        object.__init__(self)

    def visit_member_function(self):
        raise NotImplementedError()

    def visit_constructor(self):
        raise NotImplementedError()

    def visit_destructor(self):
        raise NotImplementedError()

    def visit_member_operator(self):
        raise NotImplementedError()

    def visit_casting_operator(self):
        raise NotImplementedError()

    def visit_free_function(self):
        raise NotImplementedError()

    def visit_free_operator(self):
        raise NotImplementedError()

    def visit_class_declaration(self):
        raise NotImplementedError()

    def visit_class(self):
        raise NotImplementedError()

    def visit_enumeration(self):
        raise NotImplementedError()

    def visit_namespace(self):
        raise NotImplementedError()

    def visit_typedef(self):
        raise NotImplementedError()

    def visit_variable(self):
        raise NotImplementedError()

    def visit_comment(self):
        raise NotImplementedError()



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ["inline_specifier.hpp"]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_inline_specifier(global_ns):
    inlined_funcs = global_ns.calldefs('inlined')
    assert len(inlined_funcs) != 0
    for f in inlined_funcs:
        assert f.has_inline is True

    not_inlined_funcs = global_ns.calldefs('not_inlined')
    assert len(not_inlined_funcs) != 0
    for f in not_inlined_funcs:
        assert f.has_inline is False



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)
ns = global_namespace.namespace("ns")

# Use the free_functions method to find our function
func = ns.free_function(name="myFunction")

# There are two arguments:
print(len(func.arguments))

# We can loop over them and print some information:
for arg in func.arguments:
    print(
        arg.name,
        str(arg.decl_type),
        declarations.is_std_string(arg.decl_type),
        declarations.is_reference(arg.decl_type))



# Copyright 2014-2020 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest
import platform

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


@pytest.fixture
def global_ns_fixture():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    if platform.system() == "Darwin":
        config.cflags = "-std=c++11 -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    else:
        config.cflags = "-std=c++11"
    decls = parser.parse(["test_overrides.hpp"], config)
    global_ns = declarations.get_global_namespace(decls)
    return global_ns


def test_overrides(global_ns_fixture):
    """
    Check that the override information is populated for the
    simple::goodbye function. It should contain the decl for the
    base::goodbye function.  Base::goodbye has no override so it
    will be none
    """
    base = global_ns_fixture.class_("base").member_function("goodbye")
    override_decl = global_ns_fixture.class_("simple")\
                                     .member_function("goodbye")

    assert base.overrides is None
    assert override_decl.overrides is not None
    assert override_decl.overrides == base
